@@ -5,8 +5,9 @@
 # Copyright © 2017-2021 Unrud <unrud@outlook.com>
 # Copyright © 2024-2024 Pieter Hijma <pieterhijma@users.noreply.github.com>
 # Copyright © 2024-2024 Ray <ray@react0r.com>
-# Copyright © 2024-2024 Georgiy <metallerok@gmail.com>
+# Copyright © 2024-2025 Georgiy <metallerok@gmail.com>
 # Copyright © 2024-2025 Peter Bieringer <pb@bieringer.de>
+# Copyright © 2025-2025 David Greaves <david@dgreaves.com>
 #
 # This library is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -42,6 +43,9 @@ from radicale.app.base import Access, ApplicationBase
 from radicale.item import filter as radicale_filter
 from radicale.log import logger
 
+DT_FORMAT_TIMESTAMP: str = '%Y%m%dT%H%M%SZ'
+DT_FORMAT_DATE: str = '%Y%m%d'
+
 
 def free_busy_report(base_prefix: str, path: str, xml_request: Optional[ET.Element],
                      collection: storage.BaseCollection, encoding: str,
@@ -144,7 +148,8 @@ def free_busy_report(base_prefix: str, path: str, xml_request: Optional[ET.Eleme
 
 def xml_report(base_prefix: str, path: str, xml_request: Optional[ET.Element],
                collection: storage.BaseCollection, encoding: str,
-               unlock_storage_fn: Callable[[], None]
+               unlock_storage_fn: Callable[[], None],
+               max_occurrence: int = 0,
                ) -> Tuple[int, ET.Element]:
     """Read and answer REPORT requests that return XML.
 
@@ -223,14 +228,34 @@ def xml_report(base_prefix: str, path: str, xml_request: Optional[ET.Element],
     filters = (
         root.findall(xmlutils.make_clark("C:filter")) +
         root.findall(xmlutils.make_clark("CR:filter")))
+    expand = root.find(".//" + xmlutils.make_clark("C:expand"))
+
+    # if we have expand prop we use "filter (except time range) -> expand -> filter (only time range)" approach
+    time_range_element = None
+    main_filters = []
+    for filter_ in filters:
+        # extract time-range filter for processing after main filters
+        # for expand request
+        filter_copy = copy.deepcopy(filter_)
+
+        if expand is not None:
+            for comp_filter in filter_copy.findall(".//" + xmlutils.make_clark("C:comp-filter")):
+                if comp_filter.get("name", "").upper() == "VCALENDAR":
+                    continue
+                time_range_element = comp_filter.find(xmlutils.make_clark("C:time-range"))
+                if time_range_element is not None:
+                    comp_filter.remove(time_range_element)
+
+        main_filters.append(filter_copy)
 
     # Retrieve everything required for finishing the request.
     retrieved_items = list(retrieve_items(
-        base_prefix, path, collection, hreferences, filters, multistatus))
+        base_prefix, path, collection, hreferences, main_filters, multistatus))
     collection_tag = collection.tag
     # !!! Don't access storage after this !!!
     unlock_storage_fn()
 
+    n_vevents = 0
     while retrieved_items:
         # ``item.vobject_item`` might be accessed during filtering.
         # Don't keep reference to ``item``, because VObject requires a lot of
@@ -239,7 +264,7 @@ def xml_report(base_prefix: str, path: str, xml_request: Optional[ET.Element],
         if filters and not filters_matched:
             try:
                 if not all(test_filter(collection_tag, item, filter_)
-                           for filter_ in filters):
+                           for filter_ in main_filters):
                     continue
             except ValueError as e:
                 raise ValueError("Failed to filter item %r from %r: %s" %
@@ -250,12 +275,16 @@ def xml_report(base_prefix: str, path: str, xml_request: Optional[ET.Element],
 
         found_props = []
         not_found_props = []
+        item_etag: str = ""
 
         for prop in props:
             element = ET.Element(prop.tag)
             if prop.tag == xmlutils.make_clark("D:getetag"):
-                element.text = item.etag
-                found_props.append(element)
+                if expand is not None:
+                    item_etag = item.etag
+                else:
+                    element.text = item.etag
+                    found_props.append(element)
             elif prop.tag == xmlutils.make_clark("D:getcontenttype"):
                 element.text = xmlutils.get_content_type(item, encoding)
                 found_props.append(element)
@@ -264,36 +293,59 @@ def xml_report(base_prefix: str, path: str, xml_request: Optional[ET.Element],
                     xmlutils.make_clark("CR:address-data")):
                 element.text = item.serialize()
 
-                expand = prop.find(xmlutils.make_clark("C:expand"))
-                if expand is not None and item.component_name == 'VEVENT':
-                    start = expand.get('start')
-                    end = expand.get('end')
+                if (expand is not None) and item.component_name == 'VEVENT':
+                    starts = expand.get('start')
+                    ends = expand.get('end')
 
-                    if (start is None) or (end is None):
+                    if (starts is None) or (ends is None):
                         return client.FORBIDDEN, \
                             xmlutils.webdav_error("C:expand")
 
                     start = datetime.datetime.strptime(
-                        start, '%Y%m%dT%H%M%SZ'
+                        starts, DT_FORMAT_TIMESTAMP
                     ).replace(tzinfo=datetime.timezone.utc)
                     end = datetime.datetime.strptime(
-                        end, '%Y%m%dT%H%M%SZ'
+                        ends, DT_FORMAT_TIMESTAMP
                     ).replace(tzinfo=datetime.timezone.utc)
 
-                    expanded_element = _expand(
-                        element, copy.copy(item), start, end)
+                    time_range_start = None
+                    time_range_end = None
+
+                    if time_range_element is not None:
+                        time_range_start, time_range_end = radicale_filter.parse_time_range(time_range_element)
+
+                    (expanded_element, n_vev) = _expand(
+                        element=element, item=copy.copy(item),
+                        start=start, end=end,
+                        time_range_start=time_range_start, time_range_end=time_range_end,
+                        max_occurrence=max_occurrence,
+                    )
+
+                    if n_vev == 0:
+                        logger.debug("No VEVENTs found after expansion for %r, skipping", item.href)
+                        continue
+
+                    n_vevents += n_vev
                     found_props.append(expanded_element)
                 else:
                     found_props.append(element)
+                    if hasattr(item.vobject_item, "vevent_list"):
+                        n_vevents += len(item.vobject_item.vevent_list)
+                # Avoid DoS with too many events
+                if max_occurrence and n_vevents > max_occurrence:
+                    raise ValueError("REPORT occurrences limit of {} hit"
+                                     .format(max_occurrence))
             else:
                 not_found_props.append(element)
 
         assert item.href
         uri = pathutils.unstrip_path(
             posixpath.join(collection.path, item.href))
-        multistatus.append(xml_item_response(
-            base_prefix, uri, found_props=found_props,
-            not_found_props=not_found_props, found_item=True))
+
+        if found_props or not_found_props:
+            multistatus.append(xml_item_response(
+                base_prefix, uri, found_props=found_props,
+                not_found_props=not_found_props, found_item=True, item_etag=item_etag))
 
     return client.MULTI_STATUS, multistatus
 
@@ -303,83 +355,208 @@ def _expand(
         item: radicale_item.Item,
         start: datetime.datetime,
         end: datetime.datetime,
-) -> ET.Element:
+        time_range_start: Optional[datetime.datetime] = None,
+        time_range_end: Optional[datetime.datetime] = None,
+        max_occurrence: int = 0,
+) -> Tuple[ET.Element, int]:
     vevent_component: vobject.base.Component = copy.copy(item.vobject_item)
+    logger.info("Expanding event %s", item.href)
+    logger.debug(f"Expand range: {start} to {end}")
+    logger.debug(f"Time range: {time_range_start} to {time_range_end}")
 
     # Split the vevents included in the component into one that contains the
     # recurrence information and others that contain a recurrence id to
     # override instances.
-    vevent_recurrence, vevents_overridden = _split_overridden_vevents(vevent_component)
+    base_vevent, vevents_overridden = _split_overridden_vevents(vevent_component)
 
-    dt_format = '%Y%m%dT%H%M%SZ'
+    dt_format = DT_FORMAT_TIMESTAMP
     all_day_event = False
 
-    if type(vevent_recurrence.dtstart.value) is datetime.date:
+    if type(base_vevent.dtstart.value) is datetime.date:
         # If an event comes to us with a dtstart specified as a date
         # then in the response we return the date, not datetime
-        dt_format = '%Y%m%d'
+        dt_format = DT_FORMAT_DATE
         all_day_event = True
         # In case of dates, we need to remove timezone information since
         # rruleset.between computes with datetimes without timezone information
         start = start.replace(tzinfo=None)
         end = end.replace(tzinfo=None)
+        if time_range_start is not None and time_range_end is not None:
+            time_range_start = time_range_start.replace(tzinfo=None)
+            time_range_end = time_range_end.replace(tzinfo=None)
 
     for vevent in vevents_overridden:
         _strip_single_event(vevent, dt_format)
 
     duration = None
-    if hasattr(vevent_recurrence, "dtend"):
-        duration = vevent_recurrence.dtend.value - vevent_recurrence.dtstart.value
+    if hasattr(base_vevent, "dtend"):
+        duration = base_vevent.dtend.value - base_vevent.dtstart.value
+    elif hasattr(base_vevent, "duration"):
+        try:
+            duration = base_vevent.duration.value
+            if duration.total_seconds() <= 0:
+                logger.warning("Invalid DURATION: %s", duration)
+                duration = None
+        except (AttributeError, TypeError) as e:
+            logger.warning("Failed to parse DURATION: %s", e)
+            duration = None
+
+    # Generate EXDATE to remove from expansion range
+    exdates_set: set[datetime.datetime] = set()
+    if hasattr(base_vevent, 'exdate'):
+        exdates = base_vevent.exdate.value
+        if not isinstance(exdates, list):
+            exdates = [exdates]
+
+        exdates_set = {
+            exdate.astimezone(datetime.timezone.utc) if isinstance(exdate, datetime.datetime)
+            else datetime.datetime.fromordinal(exdate.toordinal()).replace(tzinfo=None)
+            for exdate in exdates
+        }
+
+        logger.debug("EXDATE values: %s", exdates_set)
+
+    events_for_filtering = vevents_overridden
 
     rruleset = None
-    if hasattr(vevent_recurrence, 'rrule'):
-        rruleset = vevent_recurrence.getrruleset()
+    if hasattr(base_vevent, 'rrule'):
+        rruleset = base_vevent.getrruleset()
+    else:
+        # if event does not have rrule, only include base event
+        events_for_filtering = [base_vevent]
 
+    filtered_vevents = []
     if rruleset:
         # This function uses datetimes internally without timezone info for dates
-        recurrences = rruleset.between(start, end, inc=True)
+
+        # A vobject rruleset is for the event dtstart.
+        # Expanded over a given time range this will not include
+        # events which started before the time range but are still
+        # ongoing at the start of the range
+
+        # To accomodate this, reduce the start time by the duration of
+        # the event. If this introduces an extra reccurence point then
+        # that event should be included as it is still ongoing. If no
+        # extra point is generated then it was a no-op.
+        rstart = start - duration if duration and duration.total_seconds() > 0 else start
+        recurrences = rruleset.between(rstart, end, inc=True, count=max_occurrence)
+        if max_occurrence and len(recurrences) >= max_occurrence:
+            # this shouldn't be > and if it's == then assume a limit
+            # was hit and ignore that maybe some would be filtered out
+            # by EXDATE etc. This is anti-DoS, not precise limits
+            raise ValueError("REPORT occurrences limit of {} hit"
+                             .format(max_occurrence))
 
         _strip_component(vevent_component)
-        _strip_single_event(vevent_recurrence, dt_format)
+        _strip_single_event(base_vevent, dt_format)
 
-        is_component_filled: bool = False
         i_overridden = 0
 
         for recurrence_dt in recurrences:
-            recurrence_utc = recurrence_dt.astimezone(datetime.timezone.utc)
+            recurrence_utc = recurrence_dt if all_day_event else recurrence_dt.astimezone(datetime.timezone.utc)
+            logger.debug("Processing recurrence: %s (all_day_event: %s)", recurrence_utc, all_day_event)
+
+            # Apply time-range filter
+            if time_range_start is not None and time_range_end is not None:
+                dtstart = recurrence_utc
+                dtend = dtstart + duration if duration else dtstart
+                # Start includes the time, end does not
+                if not (dtstart <= time_range_end and dtend > time_range_start):
+                    logger.debug("Recurrence %s filtered out by time-range", recurrence_utc)
+                    continue
+
+            # Check exdate
+            if recurrence_utc in exdates_set:
+                logger.debug("Recurrence %s excluded by EXDATE", recurrence_utc)
+                continue
+
+            # Check for overridden instances
             i_overridden, vevent = _find_overridden(i_overridden, vevents_overridden, recurrence_utc, dt_format)
 
             if not vevent:
-                # We did not find an overridden instance, so create a new one
-                vevent = copy.deepcopy(vevent_recurrence)
+                # Create new instance from recurrence
+                vevent = base_vevent.duplicate(base_vevent)
 
                 # For all day events, the system timezone may influence the
                 # results, so use recurrence_dt
                 recurrence_id = recurrence_dt if all_day_event else recurrence_utc
+                logger.debug("Creating new VEVENT with RECURRENCE-ID: %s", recurrence_id)
+
                 vevent.recurrence_id = ContentLine(
                     name='RECURRENCE-ID',
                     value=recurrence_id, params={}
                 )
                 _convert_to_utc(vevent, 'recurrence_id', dt_format)
+                suffix = ''
+                if (dt_format == DT_FORMAT_DATE):
+                    suffix = ';VALUE=DATE'
+                else:
+                    suffix = ''
                 vevent.dtstart = ContentLine(
-                    name='DTSTART',
+                    name='DTSTART' + suffix,
                     value=recurrence_id.strftime(dt_format), params={}
                 )
-                if duration:
+                # if there is a DTEND, override it. Duration does not need changing
+                if hasattr(vevent, "dtend"):
                     vevent.dtend = ContentLine(
-                        name='DTEND',
+                        name='DTEND' + suffix,
                         value=(recurrence_id + duration).strftime(dt_format), params={}
                     )
 
-            if not is_component_filled:
-                vevent_component.vevent = vevent
-                is_component_filled = True
+            filtered_vevents.append(vevent)
+
+    # Filter overridden and non-recurring events
+    if time_range_start is not None and time_range_end is not None:
+        for vevent in events_for_filtering:
+            dtstart = vevent.dtstart.value
+
+            # Handle string values for DTSTART/DTEND
+            if isinstance(dtstart, str):
+                try:
+                    dtstart = datetime.datetime.strptime(dtstart, dt_format)
+                    if all_day_event:
+                        dtstart = dtstart.date()
+                except ValueError as e:
+                    logger.warning("Invalid DTSTART format: %s, error: %s", dtstart, e)
+                    continue
+
+            dtend = dtstart + duration if duration else dtstart
+
+            logger.debug(
+                "Filtering VEVENT with DTSTART: %s (type: %s), DTEND: %s (type: %s)",
+                dtstart, type(dtstart), dtend, type(dtend))
+
+            # Convert to datetime for comparison
+            if all_day_event and isinstance(dtstart, datetime.date) and not isinstance(dtstart, datetime.datetime):
+                dtstart = datetime.datetime.fromordinal(dtstart.toordinal()).replace(tzinfo=None)
+                dtend = datetime.datetime.fromordinal(dtend.toordinal()).replace(tzinfo=None)
+            elif not all_day_event and isinstance(dtstart, datetime.datetime) \
+                    and isinstance(dtend, datetime.datetime):
+                dtstart = dtstart.replace(tzinfo=datetime.timezone.utc)
+                dtend = dtend.replace(tzinfo=datetime.timezone.utc)
             else:
-                vevent_component.add(vevent)
+                logger.warning("Unexpected DTSTART/DTEND type: dtstart=%s, dtend=%s", type(dtstart), type(dtend))
+                continue
+
+            if dtstart < time_range_end and dtend > time_range_start:
+                if vevent not in filtered_vevents:  # Avoid duplicates
+                    logger.debug("VEVENT passed time-range filter: %s", dtstart)
+                    filtered_vevents.append(vevent)
+            else:
+                logger.debug("VEVENT filtered out: %s", dtstart)
+
+    # Rebuild component
+
+    if not filtered_vevents:
+        element.text = ""
+        return element, 0
+    else:
+        vevent_component.vevent_list = filtered_vevents
+        logger.debug("lbt: vevent_component %s", vevent_component)
 
     element.text = vevent_component.serialize()
 
-    return element
+    return element, len(filtered_vevents)
 
 
 def _convert_timezone(vevent: vobject.icalendar.RecurringComponent,
@@ -487,7 +664,7 @@ def _find_overridden(
 def xml_item_response(base_prefix: str, href: str,
                       found_props: Sequence[ET.Element] = (),
                       not_found_props: Sequence[ET.Element] = (),
-                      found_item: bool = True) -> ET.Element:
+                      found_item: bool = True, item_etag: str = "") -> ET.Element:
     response = ET.Element(xmlutils.make_clark("D:response"))
 
     href_element = ET.Element(xmlutils.make_clark("D:href"))
@@ -501,6 +678,10 @@ def xml_item_response(base_prefix: str, href: str,
                 status = ET.Element(xmlutils.make_clark("D:status"))
                 status.text = xmlutils.make_response(code)
                 prop_element = ET.Element(xmlutils.make_clark("D:prop"))
+                if (item_etag != "") and (code == 200):
+                    prop_etag = ET.Element(xmlutils.make_clark("D:getetag"))
+                    prop_etag.text = item_etag
+                    prop_element.append(prop_etag)
                 for prop in props:
                     prop_element.append(prop)
                 propstat.append(prop_element)
@@ -616,9 +797,9 @@ class ApplicationPartReport(ApplicationBase):
                 assert item.collection is not None
                 collection = item.collection
 
+            max_occurrence = self.configuration.get("reporting", "max_freebusy_occurrence")
             if xml_content is not None and \
                xml_content.tag == xmlutils.make_clark("C:free-busy-query"):
-                max_occurrence = self.configuration.get("reporting", "max_freebusy_occurrence")
                 try:
                     status, body = free_busy_report(
                         base_prefix, path, xml_content, collection, self._encoding,
@@ -633,7 +814,7 @@ class ApplicationPartReport(ApplicationBase):
                 try:
                     status, xml_answer = xml_report(
                         base_prefix, path, xml_content, collection, self._encoding,
-                        lock_stack.close)
+                        lock_stack.close, max_occurrence)
                 except ValueError as e:
                     logger.warning(
                         "Bad REPORT request on %r: %s", path, e, exc_info=True)
